/*
 * Name: main.c
 * Project: vuc-demo
 * License: GNU GPL v2 (see License.txt)
 */

//#define VUC_DBG_ON	1

//#include "../config.h"
#include "../demo.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include <stdint.h>
#include <string.h>

#include "my-avr.h"
#include "../vuc.h"

#include "main.h"
#include "uart.h"


#define BUF_SIZE	16

#define SC_STATE_WAIT	0
#define SC_STATE_REC	1
#define SC_STATE_OUT	2

u8 state, buf_i, buf[BUF_SIZE];
u8 send_dis= 0;
u16 * buf16= (u16 *)buf;
vuc_adr_t adr;

u8 scope[SCOPE_SIZE], sc_i= 0, sc_state= 0, sc_last= 0;

vuc_put_meta_t vuc_put_meta[VUC_UC_PUT_ADRS]= {
  { 4, gen_put_freq },
  { 1, gen_put_power },
  { 1, scope_put_tb },
};

vuc_get_meta_t vuc_get_meta[VUC_UC_GET_ADRS]= {
  { scope_get },
};

#ifndef VUC_DBG_ON
#  define VUC_DBG_ON	0
#endif

// protocol
#ifndef VUC_DBG_IO
#  define VUC_DBG_IO	VUC_DBG_ON
#endif


/* ------------------------------------------------------------------------- */

void uart_putn04 (u8 n)
{
  n&= 0x0f;
  uart_putc(n<=9 ? n+'0' : n-10+'A');
}

void uart_putn08 (u8 n)
{
  uart_putn04(n>>4);
  uart_putn04(n&0x0f);
}

void uart_putn16 (u16 n)
{
  uart_putn08(n>>8);
  uart_putn08(n&0xff);
}

void uart_putn32 (u32 n)
{
  uart_putn16(n>>16);
  uart_putn16(n&0xffff);
}

void dbg_vals (const char * c)
{
  u8 x;

  uart_putc(VUC_CMD_META|VUC_META_DBG);
  uart_putc(6+strlen(c)+1+vuc_put_meta[adr].no_of_bytes*3);
  uart_puts("# cmd:"); uart_puts(c);
  uart_puts(" ");
  for (x= 0; x<vuc_put_meta[adr].no_of_bytes; x++)
    { uart_puts(" "); uart_putn08(buf[x]); }
}


u8 vuc_meta (u8 subcmd, u8 len, u8 * data)
{
  uart_putc(VUC_CMD_META|VUC_META_MSG);
  uart_putc(5+2);
  uart_puts("META ");
  uart_putn08(subcmd);

  return VUC_REP_META|VUC_META_ACK;
}

void vuc_err (char * str, u16 data)
{
  uart_putc(VUC_CMD_META|VUC_META_MSG);
  uart_putc(6+strlen(str)+4+4);
  uart_puts("ERR: '");
  uart_puts(str);
  uart_puts("' 0x");
  uart_putn16(data);
}

void msg (const char * str, u16 data)
{
  uart_putc(VUC_CMD_META|VUC_META_MSG);
  uart_putc(strlen(str)+1+4);
  uart_puts(str);
  uart_putc(' ');
  uart_putn16(data);
}

void msg2 (const char * str, u16 data, u16 data2)
{
  uart_putc(VUC_CMD_META|VUC_META_MSG);
  uart_putc(strlen(str)+1+4+1+4);
  uart_puts(str);
  uart_putc(' ');
  uart_putn16(data);
  uart_putc(' ');
  uart_putn16(data2);
}

#if (VUC_DBG_ON)

void vuc_dbg_beg (u8 len)
{
  uart_putc(VUC_CMD_META|VUC_META_DBG);
  uart_putc(len);
}

void vuc_dbg_end (void)
{ }

void vuc_dbgs (void * str)
{
  uart_puts(str);
}

void vuc_dbgn4 (u8 n)
{
  uart_putn04(n);
}

void vuc_dbgn8 (u8 n)
{
  uart_putn08(n);
}

void vuc_dbgn16 (u16 n)
{
  uart_putn16(n);
}

#else

#  define vuc_dbg_beg(len) {}
#  define vuc_dbg_end() {}
#  define vuc_dbgs(str) {}
#  define vuc_dbgn4(n) {}
#  define vuc_dbgn8(n) {}
#  define vuc_dbgn16(n) {}

#endif

/* ------------------------------------------------------------------------- */

u8 start_scope (u8 * data)
{
  if (sc_state==0) {
    vuc_dbg_beg(11); vuc_dbgs("start_scope"); vuc_dbg_end();
    sc_state= 1;
  }

  return VUC_REP_META|VUC_META_ACK;
}

u8 gen_put_freq (u8 * data)
{
  u16 * buf16= (void *)data;

  vuc_dbg_beg(10+4+1+4); vuc_dbgs("timer_freq "); vuc_dbgn16(buf16[0]); vuc_dbgs(" "); vuc_dbgn16(buf16[1]); vuc_dbg_end();

  // TODO: Wait for overflow or match
  cli();
  PWM_ICR= buf16[0];
  PWM_OCR= buf16[1];
  sei();

//  start_scope(NULL);

  return VUC_REP_META|VUC_META_ACK;
}

u8 gen_put_power (u8 * data)
{
  vuc_dbg_beg(12+2); vuc_dbgs("timer_power "); vuc_dbgn8(*data); vuc_dbg_end();

  PWM_TCCR_A= ((*data&1)<<PWM_COM_1);

  return VUC_REP_META|VUC_META_ACK;
}

u8 scope_put_tb (u8 * data)
{
  vuc_dbg_beg(13+2); vuc_dbgs("scope_set_tb "); vuc_dbgn8(*data); vuc_dbg_end();

  SCOPE_OCR= *data;

  return VUC_REP_META|VUC_META_ACK;
}

u8 scope_get (u8 * data)
{
  msg(__FUNCTION__, 0);

  return VUC_REP_META|VUC_META_ACK;
}

/* ------------------------------------------------------------------------- */

void rep (i8 status)
{
  uart_putc(status);
  // TODO: error-codes
}

void new_state (u8 s)
{
  buf_i= 0;
  state= s;
//  msg("# state:", state);
}

u16 int_cnt= 0;


ISR(ADC_vect)
{
  u8 b= ADCH;
  SCOPE_TIFR|= (1<<SCOPE_OCF);
  if (sc_state==SC_STATE_WAIT) { // Wait for trigger
    if (sc_last<SCOPE_TRIGGER && b>=SCOPE_TRIGGER) {
      scope[sc_i++]= sc_last;
      scope[sc_i++]= b;
      sc_state++;
    }
    sc_last= b;
  } else
  if (sc_state==SC_STATE_REC) { // Record
    scope[sc_i++]= b;
    if (sc_i>=SCOPE_SIZE) {
      sc_i= 0;
      sc_state++; // Ready, wait for reset
    }
  }
  int_cnt++;
}

#define DELVAL 300000

#define VUC_PUT_ADRS VUC_UC_PUT_ADRS
#define VUC_GET_ADRS VUC_UC_GET_ADRS
#include "../vuc_input.h"

int main (void)
{
  u16 u;
  u8 x;
  u32 del= DELVAL;
  char * id= "vuc-demo";
  char * ver= " v" VUC_DEMO_VER " p" VUC_PROTO_VER;
//  char * ver= "vuc " VUC_PROTO_VER;

  new_state(0);

  PORTB= PORTC= PORTD= 0xff;			// activate all pull-ups
  SCOPE_PORT_ADR&= ~(1<<SCOPE_BIT_NO);		// except on ADC-Input

  DDRB= DDRC= DDRD= 0x00;			// all pins input
  LED0_DDR_ADR|= (1<<LED0_BIT_NO);		// LED-pin is output

  buf16[0]= F_CPU/8/2/PWM_DEFAULT;
  buf16[1]= F_CPU/8/2/PWM_DEFAULT/4;		// 25% PW
  gen_put_freq(buf);
  // Phase and Frequency Correct PWM, TOP=ICR
  PWM_TCCR_A= (1<<PWM_COM_1);			// Clear on Match when upcounting. Set on Match when downcounting.
  PWM_TCCR_B= (1<<PWM_WGM_3)|(1<<PWM_CS_1);	// Prescaler 8
  PWM_DDR_ADR|= (1<<PWM_BIT_NO);		// PWM-pin is output

  // ADC-trigger-timer
  SCOPE_TCCR_A= (1<<SCOPE_WGM_1);		// CTC Mode
  SCOPE_TCCR_B= SCOPE_PRESC_BITS;		// Set Prescaler
  SCOPE_OCR= F_CPU/SCOPE_PRESC_VAL/SCOPE_FREQ;	// Set Samplerate

  ADMUX= ADC_REF_BITS|ADC_IN_BITS|(1<<ADLAR);	// Ref select, input select, Result left-adjusted
  ADCSRA= ADC_PRESC_BITS|(1<<ADEN)|(1<<ADATE)|(1<<ADIE);// Enable, Auto Trigger Enable, Int when ready
  ADCSRB= (1<<ADTS1)|(1<<ADTS0);		// Auto-trigger on timer 0 compare A
  DIDR0= (1<<ADC_IN_BITS)&0x3f;			// Disable Digital Input

  uart_init(UART_BAUD_SELECT_DOUBLE_SPEED(BAUDRATE, F_CPU));
  sei();

  uart_putc(VUC_CMD_META|VUC_META_HELLO); uart_putc(VUC_PROTO_MAJOR<<4|VUC_PROTO_MINOR);
  uart_putc(VUC_CMD_META|VUC_META_ID); uart_putc(strlen(id)); uart_puts(id);
  uart_putc(VUC_CMD_META|VUC_META_MSG); uart_putc(strlen(id)+strlen(ver)); uart_puts(id); uart_puts(ver);

loop:

  if (!(del--)) {
    LED0_PORT_ADR^= (1<<LED0_BIT_NO);	// Toggle LED
    if (LED0_PORT_ADR&(1<<LED0_BIT_NO)) {
//      msg2("state/cnt:", sc_state, int_cnt);
    }
    del= DELVAL;
  }

  if (sc_state==SC_STATE_OUT) {
    uart_putc(VUC_REP_PUT|0);
    for (x=0; x<SCOPE_SIZE; x++)
      uart_putc(scope[x]);
//    sc_state++;
    sc_state= SC_STATE_WAIT;
  }

  u= uart_getc();
  if (!(u&0xff00))
    vuc_input(u&0xff);

  goto loop;

  uart_exit();
  return 0;
}

/* ------------------------------------------------------------------------- */
