
// #define VUC_DBG_ON 1
#define dbgf(fmt, args...)	printf(fmt, ##args)
//#define dbgf(fmt, args...)	{}


#include "config.h"
#include "demo.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <gtk/gtk.h>

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "anygtk.h"
#include "my-pc.h"
#include "vuc.h"

#include "rs232.h"


gboolean vuc_io_in (GIOChannel * source, GIOCondition condition, gpointer data);
gboolean vuc_io_timeout (gpointer data);


u8 scope_in (u8 * data);


RS232_HANDLE rs232_handle;

gint if_rs232_init (gchar * com_name, gint baudrate)
{
  if (RS232_OpenComport(com_name, baudrate, &rs232_handle)<0) {
    printf("Can not open com-port '%s' with %d baud.\n", com_name, baudrate);
    return 0;
  }
  g_timeout_add(5, vuc_io_timeout, NULL);
  return 1;
}


gint if_init (int argc, char * argv[])
{
  // TODO: other interface types
  return if_rs232_init(argv[2], atoi(argv[3]));
}


gint if_send_8 (u8 data)
{
//  gsize written= 0;
//  g_io_channel_write(io_ch, &data, 1, &written);
  // TODO: if (written<1) ...

  RS232_SendByte(rs232_handle, data);
  return 1;
}

gint if_send_16 (u16 data)
{
  u16 ve_data= VUC_U16_TO_VE(data);

//  gsize written= 0;
//  g_io_channel_write(io_ch, (void *)&ve_data, 2, &written);
  // TODO: if (written<2) ...

  RS232_SendByte(rs232_handle, ve_data&0xff);
  RS232_SendByte(rs232_handle, ve_data>>8);
  return 1;
}

/*
gint if_send_32 (u32 data)
{
  gsize written= 0;
  u32 ve_data= VUC_U32_TO_VE(data);

  g_io_channel_write(io_ch, (void *)&ve_data, 4, &written);
  // TODO: if (written<4) ...
  return 1;
}
*/

/*** VUC *********************************************************************/

typedef u8 (* vuc_put_f)(u8 * data);

typedef struct {
  u8 no_of_bytes;
  vuc_put_f func;
} vuc_put_meta_t;

vuc_put_meta_t vuc_put_meta[VUC_PC_PUT_ADRS]= {
  { SCOPE_SIZE, scope_in },
};


typedef u8 (* vuc_get_f)(u8 * data);

typedef struct {
  vuc_get_f func;
} vuc_get_meta_t;

vuc_get_meta_t vuc_get_meta[VUC_PC_GET_ADRS]= {
};


void print_meta (char * type, u8 len, u8 * data)
{
  printf("META %s: ", type);
  fwrite(data, len, 1, stdout);
  printf("\n");
}

u8 vuc_meta (u8 subcmd, u8 len, u8 * data)
{
  if (subcmd==VUC_META_HELLO)
    printf("META HELLO: %d.%d\n", data[0]>>4, data[0]&0x0f);

  else if (subcmd==VUC_META_ACK)
    printf("META ACK\n");
  else if (subcmd==VUC_META_NAK)
    printf("META NAK\n");

  else if (subcmd==VUC_META_ID)
    print_meta("ID", len, data);
  else if (subcmd==VUC_META_EXT)
    print_meta("EXT", len, data);
  else if (subcmd==VUC_META_MSG)
    print_meta("MSG", len, data);
  else if (subcmd==VUC_META_DBG)
    print_meta("DBG", len, data);
  else if (subcmd==VUC_META_DBG)
    print_meta("???", len, data);
}

void vuc_err (char * str, u16 data)
{
  printf("ERR: '%s' 0x%04x\n", str, data);
}

#ifdef VUC_DBG_ON

void vuc_dbg_beg (u8 len)
{ }

void vuc_dbg_end (void)
{ dbgf("\n"); }

void vuc_dbgs (void * str)
{ dbgf(str); }

void vuc_dbgn4 (u8 n)
{ dbgf("%1x", n&0x0f); }

void vuc_dbgn8 (u8 n)
{ dbgf("%02x", n); }

#else

#  define vuc_dbg_beg(len) {}
#  define vuc_dbg_end() {}
#  define vuc_dbgs(str) {}
#  define vuc_dbgn4(n) {}
#  define vuc_dbgn8(n) {}

#endif


#define IO_IN_BUF_SIZE 1024
#define VUC_IN_DATA_SIZE 256

u8 vuc_out_on= 1;

#define VUC_PUT_ADRS VUC_PC_PUT_ADRS
#define VUC_GET_ADRS VUC_PC_GET_ADRS
#include "vuc_input.h"

gint vuc_io_poll ()
{
  gsize size, read= 0;
  gchar io_in_buf[IO_IN_BUF_SIZE];
  gchar c;
#if (VUC_DBG_IO)
  GString * dbg_hex= g_string_new(NULL);
  GString * dbg_asc= g_string_new(NULL);
#endif

//  GIOStatus
  size= RS232_PollComport(rs232_handle, io_in_buf, IO_IN_BUF_SIZE);
  if (size<=0)
    return;
  while (read<size) {
    c= io_in_buf[read];
    vuc_input(io_in_buf[read++]);
#if (VUC_DBG_IO)
    g_string_append_printf(dbg_hex, " %02x", c);
    g_string_append_printf(dbg_asc, "%c", c<0x20 ? '.' : c);
#endif
  }
#if (VUC_DBG_IO)
  dbgf("IN: %s  %s\n", dbg_hex->str, dbg_asc->str);
  g_string_free(dbg_hex, 1); g_string_free(dbg_asc, 1);
#endif
}

gboolean vuc_io_in (GIOChannel * source, GIOCondition condition, gpointer data)
{
//  dbgf("vuc_io_in(%x)\n", condition);
  return vuc_io_poll(source);
}


gboolean vuc_io_timeout (gpointer data)
{
  vuc_io_poll();
  return 1;
}


/* Send some 8-bit-values to uC */
void vuc_put_8 (guint channel, guint count, ...)
{
  va_list vals;
  gint i, v;
#if (VUC_DBG_IO)
  GString * dbg_hex= g_string_new(NULL);
#endif

  // TODO: channel-numbers above 64
  if_send_8(VUC_CMD_PUT|channel);
#if (VUC_DBG_IO)
  g_string_append_printf(dbg_hex, "%02x =", VUC_CMD_PUT|channel);
#endif
  va_start(vals, count);
  for (i=0; i<count; i++) {
    // TODO: if (!if_send_16 ...
    v= va_arg(vals, int);
    if_send_8(v);
#if (VUC_DBG_IO)
    g_string_append_printf(dbg_hex, " %02x", v);
#endif
  }
  va_end(vals);

#if (VUC_DBG_IO)
  dbgf("OUT: %s\n", dbg_hex->str); g_string_free(dbg_hex, 1);
#endif
}

/* Send some 16-bit-values to uC */
void vuc_put_16 (guint channel, guint count, ...)
{
  va_list vals;
  gint i, v;
#if (VUC_DBG_IO)
  GString * dbg_hex= g_string_new(NULL);
#endif

  // TODO: channel-numbers above 64
  if_send_8(VUC_CMD_PUT|channel);
#if (VUC_DBG_IO)
  g_string_append_printf(dbg_hex, "%02x =", VUC_CMD_PUT|channel);
#endif
  va_start(vals, count);
  for (i=0; i<count; i++) {
    // TODO: if (!if_send_16 ...
    v= va_arg(vals, int);
    if_send_16(v);
#if (VUC_DBG_IO)
    g_string_append_printf(dbg_hex, " %04x", v);
#endif
  }
  va_end(vals);

#if (VUC_DBG_IO)
  dbgf("OUT: %s\n", dbg_hex->str); g_string_free(dbg_hex, 1);
#endif
}


/*** GUI *********************************************************************/

void on_window1_destroy (GtkWidget * widget, gpointer udata)
{
  gtk_main_quit();
}


/* TODO: have this auto-generated ********************************************/

GtkToggleButton * u_gen_power;
GtkAdjustment * u_gen_freq;
GtkAdjustment * u_gen_pw;

GtkDrawingArea * u_scope_area;
GtkToggleButton * u_scope_tb[8];


/*** Gen *********************************************************************/

gdouble u_gen_freq_v= 20;
gdouble u_gen_pw_v= 1;

void u_gen_set ()
{
  u16 f= (u16)(F_CPU/8/2/u_gen_freq_v);
  u16 p= (u16)(f*(u_gen_pw_v/100));
  if (p==0)
    p++;
  if (p==f)
    p--;
  vuc_put_16(0, 2, f, p);
}

void on_u_gen_power_toggled (GtkWidget * widget, gpointer udata)
{
  u8 v= gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  dbgf("Power: %d\n", v);
  vuc_put_8(1, 1, v);
}

void on_u_gen_freq_value_changed (GtkWidget * widget, gpointer udata)
{
  u_gen_freq_v= gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));
  dbgf("Freq: %04.0f\n", u_gen_freq_v);
  u_gen_set();
}

void on_u_gen_pw_value_changed (GtkWidget * widget, gpointer udata)
{
  u_gen_pw_v= gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));
  dbgf("PW: %02.0f\n", u_gen_pw_v);
  u_gen_set();
}

/*** Scope *******************************************************************/

#define SC_NORM 1024

// Timebase constants in uS/div
gint sc_timebase[8]= {
  500, 1000, 2000, 5000, 10000, 20000, 50000, 100000
};

gint sc_count= 0;
GdkPoint * sc_points= NULL;

GdkPixmap * pixmap= NULL;
GdkGC * gc= NULL;
gint width= 0;
gint height= 0;

void draw_pixmap ()
{
  GdkColormap * colormap= NULL;
  GdkColor fg;
  gint i, o, x, ox, oy, mx, my;
  GdkPoint * points= g_alloca(sizeof(GdkPoint)*max(sc_count, 2));

  if (pixmap) {
    gdk_gc_unref(gc);
    gdk_pixmap_unref(pixmap);
  }
  pixmap= gdk_pixmap_new(GTK_WIDGET(u_scope_area)->window, width, height, -1);
  gdk_draw_rectangle(pixmap, GTK_WIDGET(u_scope_area)->style->black_gc, TRUE, 0, 0, width, height);
  gc= gdk_gc_new(pixmap);
  colormap= gdk_gc_get_colormap(gc);

  ox= width*5/100;  mx= width-(2*ox);
  oy= height*5/100; my= height-(2*oy);

  if (!gdk_color_parse(SCOPE_DIV_COLOR, &fg))
    gdk_color_white(colormap, &fg);
  gdk_colormap_alloc_color(colormap, &fg, TRUE, TRUE);
  gdk_gc_set_foreground(gc, &fg);
  for (i=0; i<=8; i++) {
    o= i==4 ? ox/2 : ox;
    points[0].x= o;        points[0].y= oy+i*my/8;
    points[1].x= width-o;  points[1].y= oy+i*my/8;
    gdk_draw_lines(pixmap, gc, points, 2);
  }
  for (i=0; i<=8; i++) {
    o= i==4 ? oy/2 : oy;
    points[0].x= ox+i*mx/8; points[0].y= o;
    points[1].x= ox+i*mx/8; points[1].y= height-o;
    gdk_draw_lines(pixmap, gc, points, 2);
  }

  if (!gdk_color_parse(SCOPE_COLOR, &fg))
    gdk_color_white(colormap, &fg);
  gdk_colormap_alloc_color(colormap, &fg, TRUE, TRUE);
  gdk_gc_set_foreground(gc, &fg);
  for (x=0; x<sc_count; x++) {
    points[x].x= ox+sc_points[x].x*mx/SC_NORM;
    points[x].y= oy+sc_points[x].y*my/SC_NORM;
  }
  gdk_draw_lines(pixmap, gc, points, sc_count);
}

void draw_screen (gint x, gint y, gint w, gint h)
{
  GdkDrawable * drawable= GTK_WIDGET(u_scope_area)->window;

  gdk_draw_drawable(drawable, gc, pixmap,  x, y,  x, y,  w, h);
}

u8 scope_in (u8 * data)
{
  gint x, c, lc= 255-data[0];

//  dbgf("scope_in()\n");

  sc_count= 0;
  for (x=0; x<SCOPE_SIZE; x++) {
    c= 255-data[x];
    if (x==0 || c!=lc) {
      sc_points[sc_count].x= x*SC_NORM/SCOPE_SIZE; sc_points[sc_count].y= lc*SC_NORM/256; sc_count++;
      sc_points[sc_count].x= x*SC_NORM/SCOPE_SIZE; sc_points[sc_count].y= c*SC_NORM/256; sc_count++;
    }
    lc= c;
  }
  sc_points[sc_count].x= x*SC_NORM/SCOPE_SIZE; sc_points[sc_count].y= c*SC_NORM/256; sc_count++;
  draw_pixmap();
  draw_screen(0, 0, width, height);
  return VUC_CMD_META|VUC_META_ACK;
}

gboolean on_u_scope_area_configure_event (GtkWidget * widget, GdkEvent * event, gpointer user_data)
{
  width= widget->allocation.width;
  height= widget->allocation.height;

  draw_pixmap();
  return TRUE;
}

gboolean on_u_scope_area_expose_event (GtkWidget * widget, GdkEventExpose * event, gpointer user_data)
{
  draw_screen(event->area.x, event->area.y, event->area.width, event->area.height);
  return FALSE;
}

gboolean in_scope_tb= 0;

gboolean on_u_scope_tb_toggled (GtkWidget * widget, gpointer udata)
{
  gint i, t, v= (int)udata;

  if (in_scope_tb)
    return FALSE;

  in_scope_tb= TRUE;
  for (i=0; i<8; i++) {
    gtk_toggle_button_set_active(u_scope_tb[i], i==v);
  }
  in_scope_tb= FALSE;

  t= (F_CPU/SCOPE_PRESC_VAL) / (1000000/(sc_timebase[v]*8/SCOPE_SIZE));
  dbgf("Timebase: %d -> %5.1f mS/div -> t=%d\n", (int)udata, sc_timebase[v]/1000.0, t);
  vuc_put_8(2, 1, t);

  return FALSE;
}

/* /TODO *********************************************************************/


int main (int argc, char * argv[])
{
  gint i;
  GError * error;
  GtkBuilder * builder;
  GtkWindow * window1;
  GtkPaned * hpaned1;

  GtkScrolledWindow * dbg_scr;
  GtkAdjustment * dbg_adj;
  GtkTextView * dbg_buf;


  if (!if_init(argc, argv))
    exit(1);

  gtk_init(&argc, &argv);

  builder= gtk_builder_new();
  if (!gtk_builder_add_from_file(builder, argv[1], NULL)) {
    printf("Need input!\n");
    exit(2);
  }

  window1= GTK_WINDOW(gtk_builder_get_object(builder, "window1"));
  gtk_signal_connect(GTK_OBJECT(window1), "destroy", GTK_SIGNAL_FUNC(on_window1_destroy), NULL);

  hpaned1= GTK_PANED(gtk_builder_get_object(builder, "hpaned1"));

  dbg_scr= GTK_SCROLLED_WINDOW(gtk_builder_get_object(builder, "scrolledwindow1"));
  if (dbg_scr) {
//    dbg_adj= dbg_scr->get_vadjustment();
    dbg_buf= GTK_TEXT_VIEW(gtk_builder_get_object(builder, "textview1")); // ->get_buffer();
  }


  /* TODO: have this auto-generated ******************************************/
  u_gen_power= GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "u_gen_power"));
  gtk_signal_connect(GTK_OBJECT(u_gen_power), "toggled", GTK_SIGNAL_FUNC(on_u_gen_power_toggled), NULL);

  u_gen_freq= GTK_ADJUSTMENT(gtk_builder_get_object(builder, "u_gen_freq"));
  gtk_adjustment_set_value(u_gen_freq, 0);
  gtk_signal_connect(GTK_OBJECT(u_gen_freq), "value-changed", GTK_SIGNAL_FUNC(on_u_gen_freq_value_changed), NULL);

  u_gen_pw= GTK_ADJUSTMENT(gtk_builder_get_object(builder, "u_gen_pw"));
  gtk_adjustment_set_value(u_gen_pw, 0);
  gtk_signal_connect(GTK_OBJECT(u_gen_pw), "value-changed", GTK_SIGNAL_FUNC(on_u_gen_pw_value_changed), NULL);

  sc_points= g_malloc0(sizeof(GdkPoint)*(SCOPE_SIZE+1)*2);
  sc_points[0].x= 0;       sc_points[0].y= SC_NORM/2;
  sc_points[1].x= SC_NORM; sc_points[1].y= SC_NORM/2;
  sc_count= 2;
  u_scope_area= GTK_DRAWING_AREA(gtk_builder_get_object(builder, "u_scope_area"));
  gtk_widget_add_events(GTK_WIDGET(u_scope_area), GDK_EXPOSURE_MASK|GDK_LEAVE_NOTIFY_MASK|GDK_BUTTON_PRESS_MASK|GDK_POINTER_MOTION_MASK|GDK_POINTER_MOTION_HINT_MASK);
  gtk_signal_connect(GTK_OBJECT(u_scope_area), "configure-event", GTK_SIGNAL_FUNC(on_u_scope_area_configure_event), NULL);
  gtk_signal_connect(GTK_OBJECT(u_scope_area), "expose-event", GTK_SIGNAL_FUNC(on_u_scope_area_expose_event), NULL);

  for (i=0; i<8; i++) {
    char buf[20];
    sprintf(buf, "u_scope_tb_%d", i+1);
    u_scope_tb[i]= GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, buf));
    gtk_signal_connect(GTK_OBJECT(u_scope_tb[i]), "toggled", GTK_SIGNAL_FUNC(on_u_scope_tb_toggled), (void *)i);
  }
  /* /TODO *******************************************************************/


  gtk_widget_show(GTK_WIDGET(window1));

  gtk_main();

  return 0;
}
