#!/usr/bin/perl

# use strict;
# use warnings;


################################
#
#  CFG
#

#$PortName= '/dev/ttyS0';
#$PortName= '/dev/ttyUSB0';

#$XmlFile= './demo-dbh.glade';
my $XmlFile= './demo.glade';

#$ScopeColor= '#00ff80';


use C::Scan;

my $mk_defs= read_mk_file("config.mk");
my $CpuClk= $mk_defs->{'F_CPU'};
#$Baudrate= $mk_defs->{'BAUDRATE'};

my $c= new C::Scan 'filename' => "demo.h";
my $c_defs= $c->get('defines_no_args');

my $PortName= get_c_def('COM_PORT');
my $Baudrate= get_c_def('BAUDRATE');

my $ScopeSize= get_c_def('SCOPE_SIZE');
my $ScopeColor= get_c_def('SCOPE_COLOR');



$c= new C::Scan 'filename' => "vuc.h";
$c_defs= $c->get('defines_no_args');

$Cmd_meta= hex(get_c_def('VUC_CMD_META'));
$Cmd_put= hex(get_c_def('VUC_CMD_PUT'));
$Cmd_get= hex(get_c_def('VUC_CMD_GET'));


#printf("CpuClk:%d  PortName:'%s'  Baudrate:%d\n", $CpuClk, $PortName, $Baudrate);
#printf("ScopeSize:%d  ScopeColor:'%s'\n", $ScopeSize, $ScopeColor);
#printf("Cmd_meta:%02x  Cmd_put:%02x  Cmd_get:%02x\n", $Cmd_meta, $Cmd_put, $Cmd_get);


sub get_c_def {
  my $v= $c_defs->{$_[0]};
  $v =~ s|\s*//.*$||;
  if ($v =~ m|^"(.*)"$|) {
    $v= $1;
  }
  return $v;
}

sub read_mk_file {
  my ($file, %defs);

  open $file, $_[0] or die "Error opening ".$_[0].": ".$!;
  while (<$file>) {
    chomp $_;
    if (m/^\s*include\s*(.*)/i) {
      my $fn= $1; $fn =~ s|\$\(TOP_DIR\)|.|;
      my $hash2= read_mk_file($fn);
      @defs{keys %{$hash2}}= values %{$hash2};
    } elsif (m/^\s*([a-z_0-9]+)\s*=\s*(.*)/i) {
      $defs{$1}= $2;
    }
  }
  return \%defs;
}


################################
#
#  Project
#

my $Freq;
my $Width;

sub received {
  my ($adr, $data)= @_;

  if ($adr==0)
    { draw_scope($data); }
  else
    { debug(sprintf("Bad address: 0x%02x\n", $adr)); }
}

sub on_u_gen_freq_value_changed {
  my ($adj)= @_;
  $Freq= $adj->get_value;

  debug(sprintf("freq: %04.0f\n", $Freq));
  set_gen();
}

sub on_u_gen_pw_value_changed {
  my ($adj)= @_;
  $Width= $adj->get_value;

  debug(sprintf("width: %02.2f\n", $Width));
  set_gen();
}

sub set_gen {
  my ($f)= $CpuClk/8/2/$Freq;
  my ($p)= $f*($Width/100);
  if ($p==0)
    { $p++; }
  if ($p==$f)
    { $p--; }
  put_16(0, $f, $p);
}


################################
#
#  Debug
#

use Data::Dumper;
local $Data::Dumper::Indent = 3;

use Dumpvalue;
my $dumper = Dumpvalue->new;
$dumper->set(globPrint => 1);

sub debug {
  my ($str)= @_;

  if ($dbg_buf) {
    $dbg_buf->insert($dbg_buf->get_end_iter, $str);
    $dbg_adj->set_value(999999);
  } else {
    print($str);
  }
}

sub dbg_serial_in {
  my ($saw)= @_;
  my ($c, $dbg, $asc);

  while (length($saw)) {
    $c= substr($saw, 0, 1); $saw= substr($saw, 1);
    $dbg.= sprintf(" %02X", ord($c));
    $asc.= ord($c)<0x20 ? "." : $c;
  }
  debug(">".$dbg." '".$asc."'\n");
}


################################
#
#  IO
#

use Device::SerialPort;
use Gtk2::Helper;

$state= 0;
$in_buf= "";
$in_cnt= 0;
$cmd= 0;
$adr= 0;
$data= "";

$PortObj= new Device::SerialPort($PortName, 0) or die "Can't open $PortName: $!\n";
$PortObj->baudrate($Baudrate);
$PortObj->read_char_time(0);   # don't wait for each character
$PortObj->read_const_time(10); # 10 ms per unfulfilled "read" call


sub poll_serial {
  my ($cnt, $saw)= $PortObj->read(1024);
  my ($c, $d);

  if ($cnt<=0) {
    return;
  }

  $in_buf= substr($saw, 0, $cnt);
#  dbg_serial_in($in_buf);

  while (length($in_buf)>0) {
    $c= ord(substr($in_buf, 0, 1));
    $in_buf= substr($in_buf, 1);
    if ($state==0) {
      $cmd= $c&0xc0; $adr= $c&0x3f;
      if ($cmd==0x40) { # meta
        if ($adr==0x36) {
          $state= 1; # hello
        } elsif ($adr==0x06 || $adr==0x30 || $adr==0x31 || $adr==0x35) {
          $state= 2; # ext msg dbg id
        } else {
          debug(sprintf("META %02x\n", $adr));
        }
      } elsif ($cmd==0x80) { # put
        if ($adr==0) { $in_cnt= $ScopeSize; }
        $state= 3;
      }
    } elsif ($state==1) {
      debug(sprintf("META HELLO: %02x\n", $c));
      $state= 0;
    } elsif ($state==2) {
      $in_cnt= $c; $state= 3;
    }
    if ($state>=3) {
      if ($in_cnt>0) {
        $d= substr($in_buf, 0, $in_cnt);
        $in_buf= substr($in_buf, $in_cnt);
        $in_cnt-= length($d);
        $data.= $d;
      }
      if ($in_cnt<=0) {
        if ($cmd==0x40) {
          debug(sprintf("META %02x: '%s'\n", $adr, $data));
        } else {
          received($adr, $data);
        }
        $state= 0; $data= "";
      }
    }
  }
}

sub put_16 {
  my ($adr)= shift(@_);
  my ($output, $dbg, $count_out);

  $output= chr( $Cmd_put | $adr);
  $dbg= "Put16 ".sprintf(" %02X = ", ord($output));
  while (scalar(@_)) {
    my $w= shift(@_);
    $output.= chr($w & 0xff).chr($w >> 8);
    $dbg.= sprintf(" %04X", $w);
  }

#  debug("< ".$dbg."\n");

  $count_out= $PortObj->write($output);
  warn "write failed\n"         unless ($count_out);
  warn "write incomplete\n"     if ( $count_out != length($output) );
}


################################
#
#  GUI
#

my $scope_d    = [0,0.5, 1,0.5];

my $area       = undef;
my $pixmap     = undef;
my $gc         = undef;
my $colormap   = undef;
my $width      = undef;
my $height     = undef;

#use Gtk2 qw{-init -threads-init};
use Gtk2 qw{-init};

$builder= Gtk2::Builder->new();
$builder->add_from_file($XmlFile);

$dbg_scr= $builder->get_object("scrolledwindow1");
if ($dbg_scr) {
  $dbg_adj= $dbg_scr->get_vadjustment();
  $dbg_buf= $builder->get_object("textview1")->get_buffer();
}

$area= $builder->get_object("u_scope_area");
$area->add_events(
  [qw/exposure-mask leave-notify-mask button-press-mask pointer-motion-mask pointer-motion-hint-mask/]
);

$builder->connect_signals(undef);

$builder->get_object("u_gen_freq")->set_value(0);
$builder->get_object("u_gen_pw")->set_value(0);

$window= $builder->get_object("window1");

$window->resize(400, 300);
$builder->get_object("hpaned1")->set_position(120);

$window->show_all();


sub on_window1_destroy {
  Gtk2->main_quit();
}

my ($tag)= Gtk2::Helper->add_watch($PortObj->{FD}, in =>
  sub { poll_serial(); 1; }
);


################################
#
#  Scope
#

sub get_color {
    my ($colormap, $name) = @_;
    my $ret;

    if ($ret = $allocated_colors{$name}) {
        return $ret;
    }

    my $color= Gtk2::Gdk::Color->parse($name);
    $colormap->alloc_color($color, TRUE, TRUE);

    $allocated_colors{$name}= $color;

    return $color;
}

sub mk_pixmap {
  my ($widget, $width, $height)= @_;

  $pixmap= Gtk2::Gdk::Pixmap->new($widget->window, $width, $height, -1);
  $pixmap->draw_rectangle($widget->style->black_gc, TRUE, 0, 0, $width, $height);
  $gc       = Gtk2::Gdk::GC->new($pixmap);
  $colormap = $pixmap->get_colormap;
}

sub draw_lines {
  my($widget, $color, $lines) = @_;
  my $colormap = $widget->get_colormap;
  my $gc = $widget->{gc} || new Gtk2::Gdk::GC $widget;

  $gc->set_foreground(get_color($colormap, $color));  #  print Dumper($lines);
  $widget->draw_lines($gc, @$lines);
}

sub draw_pixmap {
  my ($x, $ox, $oy, $mx, $my, @lines);

  $ox= $width*0.05;  $mx= $width-(2*$ox);
  $oy= $height*0.05; $my= $height-(2*$oy);
  mk_pixmap($area, $width, $height);
  for ($x=0; $x<scalar(@$scope_d); $x+= 2) {
    $lines[$x]=   $ox+@$scope_d[$x]*$mx;
    $lines[$x+1]= $oy+@$scope_d[$x+1]*$my;
  }
  draw_lines($pixmap, $ScopeColor, \@lines);
}

sub on_u_scope_area_configure_event {
  my ($widget, $event)= @_;
  my ($x, @lines);

  $width= $widget->allocation->width;
  $height= $widget->allocation->height;
  draw_pixmap();
  return TRUE;
}

sub draw_screen {
  my ($x, $y, $w, $h)= @_;

  $area->window->draw_drawable(
    $area->style->fg_gc($area->state),  $pixmap,  $x, $y,  $x, $y,  $w, $h
  );
}

sub on_u_scope_area_expose_event {
  my ($widget, $event)= @_;

  draw_screen($event->area->x, $event->area->y, $event->area->width, $event->area->height);
  return FALSE;
}

sub draw_scope {
  my ($data)= @_;
  my ($l)= length($data);
  my ($c, $x, @lines);
  my ($lc)= $lc= 255-ord(substr($data, 0, 1));

#  debug(sprintf("draw_scope l=%d\n", $l));

  for ($x=0; $x<$l; $x++) {
    $c= 255-ord(substr($data, $x, 1));
    if ($x==0 || $c!=$lc) {
      push @lines, $x/$l, $lc/256;
      push @lines, $x/$l, $c/256;
    }
    $lc= $c;
  }
  push @lines, 1, $c/256;

  $scope_d= \@lines;
  draw_pixmap();
  draw_screen(0, 0, $width, $height);
}


################################
#
#  Main
#

Gtk2->main();
